/* Copyright (c) 1995 by Stanford University.  All Rights Reserved */

/***************************************************************************
   NAME
     mif_gfx.c
   PURPOSE
     Sample graphics routines for CS326
   NOTES
     
   AUTHOR
     Tsai-Yen Li (li@flamingo.Stanford.EDU)
   HISTORY
     Tsai-Yen Li - Apr 9, 1995: Created.
***************************************************************************/

#ifdef RCS
#ifndef lint
 static char rcsid[]="$Header: /u/li/Proj/mif/RCS/mif_gfx.c,v 1.4 1995/04/15 22:07:52 li Exp li $";
#endif 
#endif 
/* $Log: mif_gfx.c,v $
 * Revision 1.4  1995/04/15  22:07:52  li
 * add path interface
 *
 * Revision 1.3  1995/04/14  01:33:14  li
 * Deal with multiple robots
 *
 * Revision 1.2  1995/04/12  08:00:18  li
 * add comments and fix some bugs
 *
 * Revision 1.1  1995/04/11  21:15:40  li
 * Initial revision
 * */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/param.h>
#include <X11/keysym.h>
#include <Xm/Xm.h>
#include <Xm/MainW.h>
#include <Mrm/MrmPublic.h>
#include "planner_header.h"
#include "mif_header.h"
#include "mif_proto.h"

extern timeout_data Data;
extern drawing_data canvas_data;
extern XtAppContext app_context;
extern Widget shell,canvas,commandL;
extern int holo_path_exist,nholo_path_exist;

static timeout_data tt,tr;
static int translate=FALSE, rotate=FALSE;
static canvas_obj *chosen_obj;
static Fq *chosen_q,oldq;
static int chosen;
static obj_type chosen_type;
static Ipoint ul,lr;
static dim ws;
static int wsb_thickness;
/* To add/change a color, you need to modify the following array */
#define NUM_COLOR 9
static char *colorName[]={"Magenta1","Firebrick","ForestGreen","Orange",
			    "Black","Grey70","Grey40","SteelBlue","Black"};

#define MAX_TMP_POLY 10
static int Icur_tmp_idx;
static int Ipsize[MAX_TMP_POLY];
static Ipoint *Itmp_vert[MAX_TMP_POLY];
static int Fcur_tmp_idx;
static int Fpsize[MAX_TMP_POLY];
static Fpoint *Ftmp_vert[MAX_TMP_POLY];
#define CW 1
#define CCW -1
#define MAX_VERT 20

/*#define P2C(p1,p2) p2.x=RNDINT(p1.x*pc_scale); p2.y=canvas_data.cdim.height-RNDINT(p1.y*pc_scale);
#define C2P(p1,p2) p2.x=p1.x*cp_scale; p2.y=(canvas_data.cdim.height-p1.y)*cp_scale;*/

/***************************************************************************
  Checkout_Temp_Ipolygon
****************************************************************************/

void
Checkout_Temp_Ipolygon(Ipolygon *poly) {

  if (Icur_tmp_idx>=MAX_TMP_POLY) { 
    fprintf(logfd,"Error: Can't check out more temporary polygons\n");
    exit(-1);
  }
  if (poly->nvert+1>Ipsize[Icur_tmp_idx]) {
    if (Itmp_vert[Icur_tmp_idx]!=NULL) free(Itmp_vert[Icur_tmp_idx]);
    Itmp_vert[Icur_tmp_idx]=(Ipoint *)calloc(poly->nvert+1,sizeof(Ipoint));
    if (Itmp_vert[Icur_tmp_idx]==NULL) {
      NO_MORE_MEMORY("Checkout_Temp_Ipolygon");
    }
    Ipsize[Icur_tmp_idx]=poly->nvert+1;
  }
  poly->vert=Itmp_vert[Icur_tmp_idx];
  Icur_tmp_idx++;

}

/***************************************************************************
  Checkin_Temp_Ipolygon
****************************************************************************/

void
Checkin_Temp_Ipolygon(void) {

  Icur_tmp_idx--;

}

/***************************************************************************
  Checkout_Temp_Fpolygon
****************************************************************************/

void
Checkout_Temp_Fpolygon(Fpolygon *poly) {

  if (Fcur_tmp_idx>=MAX_TMP_POLY) { 
    fprintf(logfd,"Error: Can't check out more temporary polygons\n");
    exit(-1);
  }
  if (poly->nvert+1>Fpsize[Icur_tmp_idx]) {
    if (Ftmp_vert[Fcur_tmp_idx]!=NULL) free(Ftmp_vert[Fcur_tmp_idx]);
    Ftmp_vert[Fcur_tmp_idx]=(Fpoint *)calloc(poly->nvert+1,sizeof(Fpoint));
    if (Ftmp_vert[Fcur_tmp_idx]==NULL) {
      NO_MORE_MEMORY("Checkout_Temp_Ipolygon");
    }
    Fpsize[Fcur_tmp_idx]=poly->nvert+1;
  }
  poly->vert=Ftmp_vert[Fcur_tmp_idx];
  Fcur_tmp_idx++;

}

/***************************************************************************
  Checkin_Temp_Fpolygon
****************************************************************************/

void
Checkin_Temp_Fpolygon(void) {

  Fcur_tmp_idx--;

}

/***************************************************************************
  Conf_C2P
****************************************************************************/

void 
Conf_C2P(Fq p1,Fq *p2) {

  p2->x=p1.x*cp_scale; 
  p2->y=(canvas_data.cdim.height-p1.y)*cp_scale;

}

/***************************************************************************
  FPoint_C2P
****************************************************************************/

void 
FPoint_C2P(Fpoint p1,Fpoint *p2) {

  p2->x=p1.x*cp_scale; 
  p2->y=(canvas_data.cdim.height-p1.y)*cp_scale;

}

/***************************************************************************
  FPoint_C2P
****************************************************************************/

void 
FIPoint_C2P(Fpoint p1,Ipoint *p2) {

  p2->x=RNDINT(p1.x*cp_scale); 
  p2->y=RNDINT((canvas_data.cdim.height-p1.y)*cp_scale);

}

/***************************************************************************
  IPoint_C2P
****************************************************************************/

void 
IPoint_C2P(Ipoint p1,Ipoint *p2) {

  p2->x=RNDINT(p1.x*cp_scale); 
  p2->y=RNDINT((canvas_data.cdim.height-p1.y)*cp_scale);

}

/***************************************************************************
  IPoint_P2C
****************************************************************************/

void 
FPoint_P2C(Fpoint p1,Fpoint *p2) {

  p2->x=p1.x*pc_scale; 
  p2->y=canvas_data.cdim.height-p1.y*pc_scale;  

}

/***************************************************************************
  IPoint_P2C
****************************************************************************/

void 
IPoint_P2C(Ipoint p1,Ipoint *p2) {

  p2->x=RNDINT(p1.x*pc_scale); 
  p2->y=RNDINT(canvas_data.cdim.height-p1.y*pc_scale);  

}

/***************************************************************************
  IPoint_P2C
****************************************************************************/

void 
FIPoint_P2C(Fpoint p1,Ipoint *p2) {

  p2->x=RNDINT(p1.x*pc_scale); 
  p2->y=RNDINT(canvas_data.cdim.height-p1.y*pc_scale);  

}

/***************************************************************************
  Init_Drag_Timeout
****************************************************************************/

void 
Init_Drag_Timeout(void){ 

  tt.proc=(XtTimerCallbackProc)Track_Translation;
  tt.id=0;
  tt.delay=200L;
  tr.proc=(XtTimerCallbackProc)Track_Rotation;
  tr.id=0;
  tr.delay=50L;
}

/***************************************************************************
  Start_Rotation
****************************************************************************/

void 
Start_Rotation(int cc) {

  V3(fprintf(logfd,"tr.delay=%ld\n",tr.delay));
  XtAppAddTimeOut(app_context,tr.delay,tr.proc,(XtPointer)(long)cc);

}

/***************************************************************************
  Stop_Timeout
****************************************************************************/

void 
Stop_Timeout(timeout_data *to){

  if (to->id!=(XtIntervalId)NULL) {
    XtRemoveTimeOut(to->id);
    to->id=0;
  }
}

/***************************************************************************
  Resume_Timeout
****************************************************************************/

void 
Start_Timeout(timeout_data *to){

  to->id = XtAppAddTimeOut(app_context,to->delay,
                           to->proc,to);

}

/***************************************************************************
  close_enough
****************************************************************************/

#define EPSILON 50 /* in term of canvas scale */

int 
Close_Enough(Ipoint mouse,Ipoint orig,int *dist) {

  (*dist)=abs(orig.x-mouse.x)+abs(orig.y-mouse.y);
  return(*dist < EPSILON);

}

/***************************************************************************
  Track_Translation
****************************************************************************/

void 
Track_Translation(timeout_data *to) {

  register int i,j;
  Window root, child;
  Ipoint rootp,w,diff,canvas_p,planner_p;
  unsigned int mask;
  
  Stop_Timeout(to);
  if (!translate) return;
  /* get pointer location */
  if (XQueryPointer(canvas_data.display,canvas_data.window,&root,&child,
                    &rootp.x, &rootp.y, &w.x, &w.y, &mask) == TRUE) {
    
    /* Is it inside the work space? */
    if ((w.x<ul.x)||(w.y<ul.y)||(w.x>lr.x)||(w.y>lr.y)) {
      Start_Timeout(to);
      return;
    }
    V3(fprintf(logfd,"old point:(%d,%d)\n",w.x,w.y));

    canvas_p.x=w.x;
    canvas_p.y=w.y;
    IPoint_C2P(canvas_p,&planner_p); 
    IPoint_P2C(planner_p,&canvas_p);

    w.x=canvas_p.x;
    w.y=canvas_p.y;

    /* move the cursor to a planner grid point */
    XWarpPointer(canvas_data.display,canvas_data.window,canvas_data.window,
		 0, 0,(unsigned int)canvas_data.cdim.width, 
		 (unsigned int)canvas_data.cdim.height,w.x,w.y);
    
    V3(fprintf(logfd,"new point:(%d,%d)\n",w.x,w.y));

    diff.x=w.x-chosen_obj->conf.x;
    diff.y=w.y-chosen_obj->conf.y;

    /* chosen_obj is a global variable */
    chosen_obj->conf.x=w.x;
    chosen_obj->conf.y=w.y;

    if ((diff.x!=0)||(diff.y!=0)) {
      /* make the move */
      for(i=0;i<chosen_obj->npoly;i++) {
	for(j=0;j<=chosen_obj->cpoly[i].nvert;j++) {
	  chosen_obj->cpoly[i].vert[j].x += diff.x;
	  chosen_obj->cpoly[i].vert[j].y += diff.y;
	}
      }
      Display_Environment();
    }
  }
  Start_Timeout(to);
}

/***************************************************************************
  Track_Rotation
****************************************************************************/

void 
Track_Rotation(int sign) {

  Stop_Timeout(&tr);

  if (rotate == TRUE) {
    chosen_obj->conf.th += sign*delta_th;
    NORMALIZE(chosen_obj->conf.th,TWOPI);
    Update_Canvas_Obj_Polygon(chosen_obj);
    Display_Environment();
  } else {
    return;
  }
  
  Start_Rotation(sign);
}


/***************************************************************************
  Start_Tracking
****************************************************************************/

void 
Start_Tracking(Widget canvas, XtPointer client_data, XEvent *event) {

  register int i;
  Ipoint mouse;
  int cur_dist,min_dist=10000;
  Ipoint orig,canvas_p,planner_p;

  chosen=FALSE;

  mouse.x = event->xbutton.x;
  mouse.y = event->xbutton.y;
  
  /* the following procedure will choose the closest object */
  for(i=0;i<nrobots;i++) {
    /* Is robot init is chosen? */
    orig.x=robots[i].cobj.conf.x;
    orig.y=robots[i].cobj.conf.y;
    if (Close_Enough(mouse,orig,&cur_dist)) {
      if (cur_dist < min_dist) {
	min_dist = cur_dist;
	chosen_type = ROBOT_INIT;
	chosen_obj = &robots[i].cobj;
	chosen_q = &robots[i].pobj.init;
	chosen=TRUE;
      }
    }
    
    /* Is robot goal is chosen? */
    orig.x=robots[i].cgobj.conf.x;
    orig.y=robots[i].cgobj.conf.y;
    if (Close_Enough(mouse,orig,&cur_dist)) {
      if (cur_dist < min_dist) {
	min_dist = cur_dist;
	chosen_type = ROBOT_GOAL;
	chosen_obj = &robots[i].cgobj;
	chosen_q = &robots[i].pobj.goal;
	chosen=TRUE;
      }
    }
  }
  
  /* Is an obstacle chosen? */
  for(i=0;i<nobstacles;i++) {
    orig.x=obstacles[i].cobj.conf.x;
    orig.y=obstacles[i].cobj.conf.y;
    if (Close_Enough(mouse,orig,&cur_dist)) {
      if (min_dist>cur_dist) {
	min_dist=cur_dist;
	chosen_type=OBST;
	chosen_obj= &obstacles[i].cobj;
	chosen_q = &obstacles[i].pobj.init;
	chosen=TRUE;
      }
    }
  }

  if (!chosen) 
    return;

  /* move the pointer to the origin of the object */
  XWarpPointer(canvas_data.display,canvas_data.window,canvas_data.window,
               0,0,(unsigned int)canvas_data.cdim.width,
	       (unsigned int)canvas_data.cdim.height,
               (int)(chosen_obj->conf.x),(int)(chosen_obj->conf.y));

  /* Update the pointer and the environment */
  Display_Environment();

  /* record the old configuratio of the selected object */
  /* will be used for comparison to see if it is moved */
  if ((translate==FALSE)&&(rotate==FALSE)) {
    canvas_p.x=chosen_obj->conf.x;
    canvas_p.y=chosen_obj->conf.y;
    IPoint_C2P(canvas_p,&planner_p);
    oldq.x=planner_p.x;
    oldq.y=planner_p.y;
    oldq.th=RNDINT(chosen_obj->conf.th*rad2dth);
  }
    
  /* start appropriate timeout procedure the pressed button */
  switch(event->xbutton.button) {
  case Button1:
    translate = TRUE;
    Start_Timeout(&tt);
    break;
  
  case Button2:
    rotate = TRUE;
    Start_Rotation(CCW);
    break;

  case Button3:
    rotate = TRUE;
    Start_Rotation(CW);
    break;
  default:
    break;
  }
}

/***************************************************************************
  End_Tracking
****************************************************************************/

void 
End_Tracking(Widget canvas, XtPointer client_data, XEvent *event) {
  
  Ipoint canvas_p,planner_p;
  Iq newq;

  if (!chosen)
    return;

  if (event->xbutton.button == Button1) {
    translate = FALSE;
    Stop_Timeout(&tt);
    Conf_C2P(chosen_obj->conf,chosen_q);
  }

  if (event->xbutton.button == Button2 || event->xbutton.button == Button3) {
    rotate = FALSE;
    Stop_Timeout(&tr);
    chosen_q->th=RNDINT(chosen_obj->conf.th*rad2dth);
    NORMALIZE((chosen_q->th),dth);
  }

  /* update configurations after finishing tracking */
  if ((rotate==FALSE)&&(translate==FALSE)) {
    canvas_p.x=chosen_obj->conf.x;
    canvas_p.y=chosen_obj->conf.y;
    IPoint_C2P(canvas_p,&planner_p);
    newq.x=planner_p.x;
    newq.y=planner_p.y;
    newq.th=RNDINT(chosen_obj->conf.th*rad2dth);

    if ((newq.x!=oldq.x)||(newq.y!=oldq.y)||(newq.th!=oldq.th)) {
      if (chosen_type==ROBOT_GOAL) {
	V2(fprintf(logfd,"Goal Changed\n"));
	w_printf(commandL,"Goal Changed\n");
      } else if (chosen_type==ROBOT_INIT) {
	Update_Planner_Robots_Conf(robots);
      } else if (chosen_type==OBST) {
	V2(fprintf(logfd,"Environment Changed\n"));
	w_printf(commandL,"Environment Changed\n");
	Update_Planner_Obst_Conf(obstacles);
      }
    }
  }
  V2(fprintf(logfd,"q=(%f,%f,%f)\n",chosen_q->x,chosen_q->y,chosen_q->th));
  /* update the display */
  Display_Environment();
}


/***************************************************************************
  Hook_Tracker_Handlers
****************************************************************************/

void 
Hook_Tracker_Handlers(Widget w) {

  XtAddEventHandler(w,ButtonPressMask,FALSE,(XtEventHandler)Start_Tracking,
                    (XtPointer)NULL);

  XtAddEventHandler(w,ButtonReleaseMask,FALSE,(XtEventHandler)End_Tracking,
		    (XtPointer)NULL);
}

/***************************************************************************
  Draw_Outlined_Filled_Polygon: draw a filled polygon with boundary
  outline_c is the outline color while fill_c is the interior color.
****************************************************************************/

void 
Draw_Outlined_Filled_Polygon(Drawable d,Ipolygon *poly,int outline_c,
			     int fill_c) {

  int i;
  int num_vert;
  static int vert_size=0;
  static XPoint *vert;

  /* set up an array of XPoint for the X drawing routine */
  if (poly->nvert+1>vert_size) {
    if (vert!=NULL) free(vert);
    if ((vert=(XPoint *)calloc(poly->nvert+1,sizeof(XPoint)))==NULL) {
      NO_MORE_MEMORY("Draw_Outlined_Filled_Polygon");
    } else {
      vert_size=poly->nvert+1;
    }
  }
  num_vert=poly->nvert+1;
  for(i=0;i<num_vert;i++) {
    vert[i].x=poly->vert[i].x;
    vert[i].y=poly->vert[i].y;
  }

  /* main call to Xlib for drawing and filling a polygon */
  XFillPolygon(canvas_data.display,d,canvas_data.gc[fill_c],vert, 
	       num_vert,Convex,CoordModeOrigin);
  XDrawLines(canvas_data.display,d,canvas_data.gc[outline_c],vert, 
	     num_vert,CoordModeOrigin);

}
    
/***************************************************************************
  Draw_Outlined_Polygon: draw a outlined/empty polygon
****************************************************************************/

void 
Draw_Outlined_Polygon(Drawable d,Ipolygon *poly,int outline_c) {

  int i;
  int num_vert;
  static int vert_size=0;
  static XPoint *vert;

  /* set up an array of XPoint for the X drawing routine */
  if (poly->nvert+1>vert_size) {
    if (vert!=NULL) free(vert);
    if ((vert=(XPoint *)calloc(poly->nvert+1,sizeof(XPoint)))==NULL) {
      NO_MORE_MEMORY("Draw_Outlined_Filled_Polygon");
    } else {
      vert_size=poly->nvert+1;
    }
  }

  num_vert=poly->nvert+1;
  for(i=0;i<num_vert;i++) {
    vert[i].x=poly->vert[i].x;
    vert[i].y=poly->vert[i].y;
  }
  /* main call to Xlib for drawing a polygon */
  XDrawLines(canvas_data.display,d,canvas_data.gc[outline_c],vert, 
	     num_vert,CoordModeOrigin);
  
}

/***************************************************************************
  Draw_Filled_Polygon
****************************************************************************/

void 
Draw_Filled_Polygon(Drawable d,Ipolygon *poly,int fill_c) {

  int i;
  int num_vert;
  static int vert_size=0;
  static XPoint *vert;

  /* set up an array of XPoint for the X drawing routine */
  if (poly->nvert+1>vert_size) {
    if (vert!=NULL) free(vert);
    if ((vert=(XPoint *)calloc(poly->nvert+1,sizeof(XPoint)))==NULL) {
      NO_MORE_MEMORY("Draw_Outlined_Filled_Polygon");
    } else {
      vert_size=poly->nvert+1;
    }
  }

  num_vert=poly->nvert+1;
  for(i=0;i<num_vert;i++) {
    vert[i].x=poly->vert[i].x;
    vert[i].y=poly->vert[i].y;
  }
  /* main call to Xlib for filling a polygon */
  XFillPolygon(canvas_data.display,d,canvas_data.gc[fill_c],vert,num_vert,
               Convex,CoordModeOrigin);
  
}

/***************************************************************************
  Get_Pixel: get the pixel for the given color
****************************************************************************/

unsigned long 
Get_Pixel(char *color_name) {
  
  XColor pcolor, ignore;

  if (XAllocNamedColor(canvas_data.display,canvas_data.colormap,color_name,
                       &pcolor,&ignore)) 
    return pcolor.pixel;
  else {
    fprintf(logfd,"Error: couldn't allocate color %s, allocated %x\n",
	    color_name,
	    (unsigned int)BlackPixel(canvas_data.display, canvas_data.screen));
    return BlackPixel(canvas_data.display, canvas_data.screen);
  }
}

/***************************************************************************
  Init_Canvas: Allocate and initialize GC. Each GC is for a color.
  Create a pixel map buffer.
****************************************************************************/

void 
Init_Canvas(Widget w) {

  int i;
  unsigned long bg_pixel;

  canvas_data.widget = w;
  canvas_data.display = XtDisplay(w);
  canvas_data.window = XtWindow(w);
  canvas_data.screen = DefaultScreen(canvas_data.display);
  canvas_data.colormap = 
    DefaultColormap(canvas_data.display, canvas_data.screen);
  canvas_data.depth = DefaultDepth(canvas_data.display, canvas_data.screen);
  canvas_data.widget=w;
  /* The following line works only on some platforms */
  /*XtVaGetValues(canvas_data.widget,XmNwidth, &canvas_data.cdim.width,
                XmNheight, &canvas_data.cdim.height,NULL);*/
  /* on some machines like SUN and SGI's, we need to hardcode it 
     with the following two lines */
  canvas_data.cdim.height=640;
  canvas_data.cdim.width=640;

  /* allocate the color and GC that are going to be used */
  XtVaGetValues(w,XmNbackground, &bg_pixel,NULL);

  if ((canvas_data.gc=(GC *) calloc(NUM_COLOR,sizeof(GC)))==NULL) {
    NO_MORE_MEMORY("Init_Canvas");
  }
  for (i=0; i<NUM_COLOR; i++) {
    canvas_data.gc[i] = XCreateGC(canvas_data.display,canvas_data.window,
				  0L, NULL);
    XSetForeground(canvas_data.display,canvas_data.gc[i],
                   Get_Pixel(colorName[i]));
    XSetFunction(canvas_data.display,canvas_data.gc[i],GXcopy);
  } 
  
  wsb_thickness=canvas_data.cdim.width/(MAX(dx,dy));
  XSetLineAttributes(canvas_data.display,canvas_data.gc[thick_black], 
                     wsb_thickness,LineSolid,CapRound,JoinRound);

  canvas_data.gc[bg]=XCreateGC(canvas_data.display,canvas_data.window,0L,NULL);
  XSetForeground(canvas_data.display,canvas_data.gc[bg],bg_pixel);
  canvas_data.pmap = 
    XCreatePixmap(canvas_data.display,RootWindow(canvas_data.display,
						 canvas_data.screen),
		  canvas_data.cdim.width,canvas_data.cdim.height,
		  canvas_data.depth);

}
  
/****************************************************************************
  Convert_P2C_Polygon
****************************************************************************/

void 
Convert_P2C_Polygon(Fpolygon poly1,Ipolygon *poly2) {

  int i;
  
  poly2->nvert=poly1.nvert;

  for(i=0;i<=poly1.nvert;i++) {
    FIPoint_P2C(poly1.vert[i],&(poly2->vert[i]));
  }

}

/****************************************************************************
  Transform_Planner_Polygon: transform p1 to p2 according to orig and theta.
****************************************************************************/

void 
Transform_Planner_Polygon(Fpoint orig,float theta,Fpolygon p1,
			  Fpolygon *p2){
  
  register int i;
  float cth=cos(theta),sth=sin(theta),x,y;

  /* p1 and p2 could be the same polygon */
  for(i=0;i<=p1.nvert;i++) {
    x=(p1.vert[i].x*cth)-((p1.vert[i].y)*sth)+orig.x;
    y=(p1.vert[i].x*sth)+((p1.vert[i].y)*cth)+orig.y;
    p2->vert[i].x=x;
    p2->vert[i].y=y;
  }
}

/****************************************************************************
  Update_Planner_Robots_Conf: update the robot init and goal configurations 
  for the planner from the configuration of its canvas data
****************************************************************************/

void 
Update_Planner_Robots_Conf(robot_data *robots){

  register int i,j;
  Fpoint corig,porig;
  robot_data *robot;

  for(i=0;i<nrobots;i++) {
    robot = robots+i;
    /* update the initial configuration */
    corig.x=robot->cobj.conf.x;
    corig.y=robot->cobj.conf.y;
    FPoint_C2P(corig,&porig);
    robot->pobj.init.x=porig.x;
    robot->pobj.init.y=porig.y;
    robot->pobj.init.th=robot->cobj.conf.th*rad2dth;
    
    for(j=0;j<robot->pobj.npoly;j++) {
      robot->pobj.poly[j].nvert=robot->pobj.orig_poly[j].nvert;
      Transform_Planner_Polygon(porig,robot->cobj.conf.th,
				robot->pobj.orig_poly[j],&robot->pobj.poly[j]);
    }
    /* update the goal configuration */
    corig.x=robot->cgobj.conf.x;
    corig.y=robot->cgobj.conf.y;
    FPoint_C2P(corig,&porig);
    robot->pobj.goal.x=porig.x;
    robot->pobj.goal.y=porig.y;
    robot->pobj.goal.th=robot->cgobj.conf.th*rad2dth;
  }

}

/****************************************************************************
  Update_Canvas_Obj_Polygon: update a canvas object (robot/obstacle) from its
  canvas configuration
****************************************************************************/

void 
Update_Canvas_Obj_Polygon(canvas_obj *obj) {

  register int i;
  Fpoint corig,porig;
  Fpolygon poly;

  /* convert the configuration to planner scale */
  corig.x=obj->conf.x;
  corig.y=obj->conf.y;
  FPoint_C2P(corig,&porig);

  for(i=0;i<obj->npoly;i++) {
    poly.nvert=obj->cpoly[i].nvert;
    Checkout_Temp_Fpolygon(&poly);
    /* obtain the canvas polygon from its original polygon in planner scale */
    Transform_Planner_Polygon(porig,obj->conf.th,(*(obj->orig_poly_ptr))[i],
			      &poly);
    Convert_P2C_Polygon(poly,&obj->cpoly[i]);
    Checkin_Temp_Fpolygon();
  }

}

/****************************************************************************
  Update_Canvas_Robots_Conf: update the canvas configuration and polygon
  from the planner configuration
****************************************************************************/

void 
Update_Canvas_Robots_Conf(robot_data *robots){

  register int i,j;
  Fpoint porig,corig;
  Fpolygon poly;
  robot_data *robot;

  for(i=0;i<nrobots;i++) {
    robot = robots+i;
    /* update the initial canvas configuration */
    porig.x=robot->pobj.init.x;
    porig.y=robot->pobj.init.y;
    FPoint_P2C(porig,&corig);
    robot->cobj.conf.x=corig.x;
    robot->cobj.conf.y=corig.y;
    robot->cobj.conf.th=robot->pobj.init.th*dth2rad;
    
    /* update the initial robot polygon for the canvas */
    for (j=0;j<robot->pobj.npoly;j++) {
      poly.nvert=robot->pobj.orig_poly[j].nvert;
      Checkout_Temp_Fpolygon(&poly);
      Transform_Planner_Polygon(porig,robot->cobj.conf.th,
				robot->pobj.orig_poly[j],&poly);
      robot->cobj.cpoly[j].nvert=poly.nvert;
      Convert_P2C_Polygon(poly,&robot->cobj.cpoly[j]);
      Checkin_Temp_Fpolygon();
    }
    
    /* update the goal canvas configuration */
    porig.x=robot->pobj.goal.x;
    porig.y=robot->pobj.goal.y;
    FPoint_P2C(porig,&corig);
    robot->cgobj.conf.x=corig.x;
    robot->cgobj.conf.y=corig.y;
    robot->cgobj.conf.th=robot->pobj.goal.th*dth2rad;
    
    /* update the goal robot polygon for the canvas */
    for (j=0;j<robot->pobj.npoly;j++) {
      poly.nvert=robot->pobj.orig_poly[j].nvert;
      Transform_Planner_Polygon(porig,robot->cgobj.conf.th,
				robot->pobj.orig_poly[j],&poly);
      
      robot->cgobj.cpoly[j].nvert=poly.nvert;
      Convert_P2C_Polygon(poly,&robot->cgobj.cpoly[j]);
    }
  }
}

/****************************************************************************
  Update_Canvas_Obst_Conf: update the canvas configuration and polygon
  from the planner configuration
****************************************************************************/

void 
Update_Canvas_Obst_Conf(obst_data *obsts){

  register int i,j;
  Fpoint corig,porig;
  Fpolygon poly;

  for(i=0;i<nobstacles;i++) {

    /* update the obstacle canvas configuration */
    porig.x=obsts[i].pobj.init.x;
    porig.y=obsts[i].pobj.init.y;
    FPoint_P2C(porig,&corig);
    obsts[i].cobj.conf.x=corig.x;
    obsts[i].cobj.conf.y=corig.y;
    obsts[i].cobj.conf.th=obsts[i].pobj.init.th*dth2rad;
    
    /* update the obstcle polygons for the canvas */
    for(j=0;j<obsts[i].pobj.npoly;j++) {
      poly.nvert=obsts[i].pobj.orig_poly[j].nvert;
      Checkout_Temp_Fpolygon(&poly);
      Transform_Planner_Polygon(porig,obsts[i].cobj.conf.th,
				obsts[i].pobj.orig_poly[j],&poly);
      
      obsts[i].cobj.cpoly[j].nvert=poly.nvert;
      Convert_P2C_Polygon(poly,&obsts[i].cobj.cpoly[j]);
      Checkin_Temp_Fpolygon();
    }
  }
}

/****************************************************************************
  Update_Planner_One_Obst_Conf: update the planner obstacle configuration 
  from its canvas data (For only one obstacle)
****************************************************************************/

void 
Update_Planner_One_Obst_Conf(obst_data *obst){

  register int i;
  Fpoint corig,porig;

  /* update the obstacle configuration for the planner */
  corig.x=obst->cobj.conf.x;
  corig.y=obst->cobj.conf.y;
  FPoint_C2P(corig,&porig);
  obst->pobj.init.x=porig.x;
  obst->pobj.init.y=porig.y;
  obst->pobj.init.th=obst->cobj.conf.th*rad2dth;
  NORMALIZE(obst->pobj.init.th,dth);
  
  /* update the obstacle polygon for the planner */
  for(i=0;i<obst->pobj.npoly;i++) {
    obst->pobj.poly[i].nvert=obst->pobj.orig_poly[i].nvert;
    Transform_Planner_Polygon(porig,obst->cobj.conf.th,
			      obst->pobj.orig_poly[i],&obst->pobj.poly[i]);
  }

}

/****************************************************************************
  Update_Planner_Obst_Conf: Update all obstacle data for the planner from its
  canvas data
****************************************************************************/

void 
Update_Planner_Obst_Conf(obst_data *obsts){

  register int i;

  for(i=0;i<nobstacles;i++) {
    Update_Planner_One_Obst_Conf(obsts+i);
  }

}

/***************************************************************************
  Display_Obstacles: Draw all obstacles on the buffered pmap
****************************************************************************/

void 
Display_Obstacles(void) {

  register int i,j;

  for(i=0;i<nobstacles;i++) {
    for(j=0;j<obstacles[i].cobj.npoly;j++) {
      Draw_Outlined_Filled_Polygon(canvas_data.pmap,
				   &obstacles[i].cobj.cpoly[j],
				   obstacles[i].cobj.outline_c,
				   obstacles[i].cobj.fill_c);
    }
  }
}

/***************************************************************************
  Display_One_Robot_On_Screen: Draw a robot on the screen 
  with the given colors
****************************************************************************/

/*void 
Display_One_Robot_On_Screen(int npoly,Ipolygon *poly,int oc,int fc,int flush) {

  register int i;

  for (i=0;i<npoly;i++) {
    Draw_Outlined_Polygon(canvas_data.pmap,poly+i,fc);
    Draw_Outlined_Polygon(canvas_data.window,poly+i,fc);
  }

  if (flush) {
    XFlush(canvas_data.display);
  }

}*/

/***************************************************************************
  Display_One_Robot: Draw one robot on the screen with the given colors
****************************************************************************/

void 
Display_One_Robot(int npoly,Ipolygon *poly,int oc,int fc,int flush) {

  register int i;

  for (i=0;i<npoly;i++) {
    Draw_Outlined_Polygon(canvas_data.pmap,poly+i,fc);
    Draw_Outlined_Polygon(canvas_data.window,poly+i,fc);
  }

  if (flush) {
    XFlush(canvas_data.display);
  }
    
}

/***************************************************************************
  Display_One_Robot_On_Screen: Draw one robot on the screen (and buffer too)
  with the given colors
****************************************************************************/

void 
Display_One_Robot_On_Screen(int npoly,Ipolygon *poly,int oc,int fc,int flush) {

  register int i;

  for (i=0;i<npoly;i++) {
    Draw_Outlined_Polygon(canvas_data.pmap,poly+i,fc);
    Draw_Outlined_Polygon(canvas_data.window,poly+i,fc);
  }
  if (flush) {
    XFlush(canvas_data.display);
  }    
}

/***************************************************************************
  Display_One_Robot_On_Buffer: Draw one robot on the buffered pmap with 
  the given colors
****************************************************************************/

void 
Display_One_Robot_On_Buffer(int npoly,Ipolygon *poly,int oc,int fc,int flush) {

  register int i;

  for (i=0;i<npoly;i++) {
    Draw_Outlined_Filled_Polygon(canvas_data.pmap,poly+i,oc,fc);
  }
  if (flush) {
    XFlush(canvas_data.display);
  }    

}

/***************************************************************************
  Display_Robots: Draw robots in its init and goal configurations 
  of the robots on the buffered pmap
****************************************************************************/

void 
Display_Robots(void) {

  register int i;

  for (i=0;i<nrobots;i++) {
    Display_One_Robot_On_Buffer(robots[i].cobj.npoly,robots[i].cobj.cpoly,
				robots[i].cobj.outline_c,
				robots[i].cobj.fill_c,FALSE);
    Display_One_Robot_On_Buffer(robots[i].cobj.npoly,robots[i].cgobj.cpoly,
				robots[i].cgobj.outline_c,
				robots[i].cgobj.fill_c,FALSE);
  }
}

/***************************************************************************
  Get_Path_Step_Polygon: For a given configuration q of a robot,
  compute and return its canvas polygon.
****************************************************************************/

void 
Get_Path_Step_Polygon(Fq q,robot_data *robot,Ipolygon *ipoly) {

  register int i;
  Fpoint porig;
  float th;
  Fpolygon fpoly;

  porig.x=q.x;
  porig.y=q.y;
  th=q.th*dth2rad;

  for (i=0;i<robot->cobj.npoly;i++) {
    fpoly.nvert=robot->cobj.cpoly[i].nvert;
    Checkout_Temp_Fpolygon(&fpoly);
    Transform_Planner_Polygon(porig,th,robot->pobj.orig_poly[i],&fpoly);
    Convert_P2C_Polygon(fpoly,&ipoly[i]);
    Checkin_Temp_Fpolygon();
  }
}

/***************************************************************************
  Display_Path_Step: Draw a robot on screen for a given q and color c
****************************************************************************/

void 
Display_Path_Step(Fq q,robot_data *robot,int c) {

  Get_Path_Step_Polygon(q,robot,robot->cobj.itpoly);
  Display_One_Robot_On_Screen(robot->cobj.npoly,robot->cobj.itpoly,black,c,
			      FALSE);

}

/***************************************************************************
  Display_Robots_Path_Step_On_Buffer: Draw all robots on the buffer
  for a given q and color c
****************************************************************************/

void 
Display_Robots_Path_Step_On_Buffer(int path_idx,path_type ptype) {

  int i;

  if (ptype==HOLO) {
    for (i=0;i<nrobots;i++) {
      if (robots[i].pobj.holo_path.len > path_idx) {
	Get_Path_Step_Polygon(robots[i].pobj.holo_path.path[path_idx],
			      robots+i,robots[i].cobj.itpoly);
	Display_One_Robot_On_Buffer(robots[i].cobj.npoly,
				    robots[i].cobj.itpoly,black,
				    robots[i].animate_color,FALSE);
      }
    }
  } else {
    for (i=0;i<nrobots;i++) {
      if (robots[i].pobj.nholo_path.len > path_idx) {
	Get_Path_Step_Polygon(robots[i].pobj.nholo_path.path[path_idx],
			      robots+i,robots[i].cobj.itpoly);
	Display_One_Robot_On_Buffer(robots[i].cobj.npoly,
				    robots[i].cobj.itpoly,black,
				    robots[i].animate_color,FALSE);
      }
    }
  }

}

/***************************************************************************
  Display_Robots_Path_Step: Draw all robots on screen (and the buffer) 
  for a given q and color c
****************************************************************************/

void 
Display_Robots_Path_Step(int path_idx,path_type ptype) {
  
  int i;

  if (ptype==HOLO) {
    for (i=0;i<nrobots;i++) {
      if (robots[i].pobj.holo_path.len > path_idx) {
	Get_Path_Step_Polygon(robots[i].pobj.holo_path.path[path_idx],
			      robots+i,robots[i].cobj.itpoly);
	Display_One_Robot_On_Screen(robots[i].cobj.npoly,
				    robots[i].cobj.itpoly,black,
				    robots[i].trace_color,FALSE);
      }
    }
  } else {
    for (i=0;i<nrobots;i++) {
      if (robots[i].pobj.nholo_path.len > path_idx) {
	Get_Path_Step_Polygon(robots[i].pobj.nholo_path.path[path_idx],
			      robots+i,robots[i].cobj.itpoly);
	Display_One_Robot_On_Screen(robots[i].cobj.npoly,
				    robots[i].cobj.itpoly,black,
				    robots[i].trace_color,FALSE);
      }
    }
  }
  XFlush(canvas_data.display);
}

/***************************************************************************
  Draw_Workspace_Boundary
****************************************************************************/

void 
Draw_Workspace_Boundary(int thickness) {

  XDrawRectangle(canvas_data.display,canvas_data.pmap,
		 canvas_data.gc[thick_black],
		 thickness/2,thickness/2,ws.width-thickness,
		 ws.height-thickness);

}

/***************************************************************************
  Refresh_Environment: copy from the buffered pmap to the screen
****************************************************************************/

void 
Refresh_Environment(void) {
    
  XCopyArea(canvas_data.display,canvas_data.pmap,canvas_data.window,
            canvas_data.gc[bg],0,0,canvas_data.cdim.width,
	    canvas_data.cdim.height,0,0);

  XFlush(canvas_data.display);
}


/****************************************************************************
  Display_Environment_WO_Flush: Draw all objects in the env 
  onto the buffered pmap 
****************************************************************************/

void 
Display_Environment_WO_Flush(void) {

  /* clear the pixel map */
  XFillRectangle(canvas_data.display, canvas_data.pmap, canvas_data.gc[bg], 
                 0,0,canvas_data.cdim.width, canvas_data.cdim.height );
 
  Draw_Workspace_Boundary(wsb_thickness);
  Display_Obstacles();
  Display_Robots();

}

/****************************************************************************
  Display_Environment: Draw all objects in the env onto the screen 
****************************************************************************/

void 
Display_Environment(void) {

  Display_Environment_WO_Flush();

  XCopyArea(canvas_data.display,canvas_data.pmap,canvas_data.window,
            canvas_data.gc[bg],0,0,canvas_data.cdim.width,
	    canvas_data.cdim.height,0,0);

  XFlush(canvas_data.display);

}

/****************************************************************************
  Init_Canvas_Objects: Initialize the canvas polygons and configurations 
  from the planner/input data
****************************************************************************/

void 
Init_Canvas_Objects(void) {

  Update_Canvas_Robots_Conf(robots);
  Update_Canvas_Obst_Conf(obstacles);
  Update_Planner_Robots_Conf(robots);
  Update_Planner_Obst_Conf(obstacles);

}

/****************************************************************************
  Init_Scale: Initialize the scaling factor between the planner and the canvas
****************************************************************************/

void 
Init_Scale(void) {

  int dxy;

  dxy=MAX(dx,dy);
  pc_scale = (float)canvas_data.cdim.width/(float)dxy;
  cp_scale = (float)dxy/(float)canvas_data.cdim.width;
  ws.width=pc_scale*dx;
  ws.height=pc_scale*dy;
  ul.x=0;
  ul.y=canvas_data.cdim.height-ws.height;
  lr.x=ws.width;
  lr.y=canvas_data.cdim.height;

}

/****************************************************************************
  Init_Graphics: Initialize misc stuff for graphics 
****************************************************************************/

void 
Init_Graphics(void) {

  /* initialize the event handers */
  Hook_Tracker_Handlers(canvas);
  /* initialize the time out procedures for dragging and animation */
  Init_Timeout_Procedure();
  /* initialize the canvas data structure and gc's */
  Init_Canvas(canvas);
  /* Initialize the scaling factor between the planner and the canvas */
  Init_Scale();
  Init_Canvas_Objects();

}
