
/* defining constants */

/* for menu options */
#define k_load_robot 1
#define k_load_obst 2
#define k_load_path 3
#define k_save_robot 4
#define k_save_obst 5
#define k_save_path 6
#define k_smooth 1
#define k_debug2 2
#define k_debug3 3
  
#define MAX_GC 30
#define MAX_STR 256

/* defining data structures */
typedef enum _obj_type {ROBOT_INIT,ROBOT_GOAL,OBST} obj_type;
typedef enum _path_type {HOLO,NONHOLO} path_type;
/* to add a color, you need to modify this data type */
typedef enum _color {bg,red,green,orange,black,grey50,grey40,blue,thick_black} color;

typedef struct _dim {
  int width;
  int height;
} dim;

typedef struct _timeout_data{
  XtIntervalId id; /* returned id */
  unsigned long delay; /* interval between two calls */
  XtTimerCallbackProc proc; /* timeout procedure pointer */
} timeout_data;

typedef struct _drawing_data {
  Widget widget; /* canvas widget id */
  Display *display; 
  Window window;
  unsigned int screen;
  Pixmap pmap; /* buffered pixmap */
  Colormap colormap;
  unsigned int depth;
  dim cdim; /* dimension of the canvas */
  GC *gc; /* graphics contexts */
} drawing_data;

typedef struct _canvas_obj {
  obj_type type; 
  int npoly; /* number of polygons */
  Ipolygon *cpoly; /* polygons in canvas coord system */
  Fpolygon **orig_poly_ptr; /* pointers to the original polygons for planner */
  Fpolygon *ftpoly; /* temporary floating-point polygon */
  Ipolygon *itpoly; /* temporary integer polygon */
  Fq conf; /* configuration in canvas coord system */
  int outline_c; /* outline color */
  int fill_c;    /* interior color */
} canvas_obj;

typedef struct _robot_data {
  int id;
  canvas_obj cobj; /* for robot in its initial conf */
  canvas_obj cgobj;/* for robot in its goal conf */
  planner_obj pobj;/* for robot in the planner coord system */
  int animate_color;
  int trace_color;
} robot_data;

typedef struct _obst_data {
  int id;
  canvas_obj cobj; /* for obstacle in the canvas coord system */
  planner_obj pobj;/* for obstacle in the planner coord system */
} obst_data;


/* declaring external global variables */

extern drawing_data canvas_data;
extern timeout_data t_data;
extern holo_path_exist,nholo_path_exist;
extern robot_data *robots;
extern obst_data *obstacles;
extern objpath *curpath;
extern int debug_smooth;
extern double dt;
extern float wscale;
extern FILE *logfd;
